"""
The common code for checking message specification. 
This can be used in the tests to verify messages.

These will remain same between different versions of
message specification.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import re
from distutils.version import LooseVersion as V

import nose.tools as nt

from traitlets import (
    HasTraits, TraitError, Unicode, Dict, observe
)

string_types = (str, type(u''))

#-----------------------------------------------------------------------------
# Message Spec References
#-----------------------------------------------------------------------------

class Reference(HasTraits):

    """
    Base class for message spec specification testing.

    This class is the core of the message specification test.  The
    idea is that child classes implement trait attributes for each
    message keys, so that message keys can be tested against these
    traits using :meth:`check` method.

    """

    def check(self, d):
        """validate a dict against our traits"""
        for key in self.trait_names():
            nt.assert_in(key, d)
            # FIXME: always allow None, probably not a good idea
            if d[key] is None:
                continue
            try:
                setattr(self, key, d[key])
            except TraitError as e:
                assert False, str(e)


class Version(Unicode):
    def __init__(self, *args, **kwargs):
        self.min = kwargs.pop('min', None)
        self.max = kwargs.pop('max', None)
        kwargs['default_value'] = self.min
        super(Version, self).__init__(*args, **kwargs)
    
    def validate(self, obj, value):
        if self.min and V(value) < V(self.min):
            raise TraitError("bad version: %s < %s" % (value, self.min))
        if self.max and (V(value) > V(self.max)):
            raise TraitError("bad version: %s > %s" % (value, self.max))

mime_pat = re.compile(r'^[\w\-\+\.]+/[\w\-\+\.]+$')

class MimeBundle(Reference):
    metadata = Dict()
    data = Dict()

    @observe('data')
    def _data_changed(self, change):
        for k,v in change['new'].items():
            assert mime_pat.match(k)
            nt.assert_is_instance(v, string_types)



import sys
import io
import json
import unittest
import warnings
from collections import OrderedDict
from typing import Optional, Dict, Any
from contextlib import contextmanager


__all__ = ['JSONTestResult', 'JSONTestRunner']


class JSONTestResult(unittest.TestResult):
	def __init__(self, stream, failfast, buffer, tb_locals):
		super().__init__(stream)
		self.stream = stream
		self.failfast = failfast
		self.buffer = buffer
		self.tb_locals = tb_locals
	
	def test_run(self, test):
		self.startTestRun()
		try:
			test(self)
		finally:
			self.stopTestRun()
	
	def result_to_dict(self, type_: str, test: unittest.TestCase, err: Optional[Any]=None) -> Dict[str, Optional[str]]:
		msg = None
		if isinstance(err, tuple) and len(err) == 3:
			msg = self._exc_info_to_string(err, test)
		elif err:
			msg = str(err)
		
		return OrderedDict([
			('type', type_),
			('id', test.id()),
			('desc', test.shortDescription()),  # May be None
			('msg', msg),
		])
	
	def write_result(self, typ_: str, test: unittest.TestCase, err: Optional[Any]=None):
		json.dump(self.result_to_dict(typ_, test, err), self.stream, separators=(',', ':'))
		self.stream.write('\n')
	
	def addSuccess(self, test):
		super().addSuccess(test)
		self.write_result('success', test)
	
	def addExpectedFailure(self, test, err):
		super().addExpectedFailure(test, err)
		self.write_result('expected_failure', test, err)
	
	def addFailure(self, test, err):
		super().addFailure(test, err)
		self.write_result('failure', test, err)
	
	def addError(self, test, err):
		super().addError(test, err)
		self.write_result('error', test, err)
	
	def addUnexpectedSuccess(self, test):
		super().addUnexpectedSuccess(test)
		self.write_result('unexpected_success', test)
	
	def addSkip(self, test, reason):
		super().addSkip(test, reason)
		self.write_result('skip', test, reason)
	
	def addSubTest(self, test, subtest, err):
		super().addSubTest(test, subtest, err)
		if err is None:
			self.write_result('success', subtest)
		elif issubclass(err[0], test.failureException):
			self.write_result('failure', subtest, err)
		else:
			self.write_result('error', subtest, err)


class JSONTestRunner:
	"""TODO"""
	def __init__(
		self,
		stream: io.TextIOBase=None,
		failfast: bool=False,
		buffer: bool=False,
		warnings: Optional[str]=None,
		*,
		tb_locals: bool=False
	):
		"""Construct a JSONTestRunner."""
		self.stream = sys.stdout if stream is None else stream
		self.failfast = failfast
		self.buffer = buffer
		self.tb_locals = tb_locals
		self.warnings = warnings
	
	@contextmanager
	def filter_warnings(self):
		"""Install own warnings filter for the context"""
		with warnings.catch_warnings():
			if self.warnings:
				warnings.simplefilter(self.warnings)
				if self.warnings in ['default', 'always']:
					warnings.filterwarnings('module', category=DeprecationWarning, message=r'Please use assert\w+ instead.')
			yield
	
	def run(self, test):
		"""Run the given test case or test suite."""
		result = JSONTestResult(self.stream, self.failfast, self.buffer, self.tb_locals)
		unittest.registerResult(result)
		with self.filter_warnings():
			result.test_run(test)
		return result


if __name__ == '__main__':
	module = sys.argv[1]
	sys.argv[1:] = sys.argv[2:]
	unittest.main(module, testRunner=JSONTestRunner)



# Test manually via
# IR_KERNEL_NAME=ir python3 -m test_ir -k some_test

import os
import sys
from pathlib import Path

HERE = Path(__file__).parent
sys.path.insert(0, str(HERE / 'jkt'))

import unittest
import jupyter_kernel_test as jkt

from jupyter_client.manager import start_new_kernel
from jupyter_kernel_test.messagespec import validate_message


without_rich_display = '''\
options(jupyter.rich_display = FALSE)
{}
options(jupyter.rich_display = TRUE)
'''
#this will not work!
#withr::with_options(list(jupyter.rich_display = FALSE), {})

TIMEOUT = 15


class IRkernelTests(jkt.KernelTests):
    kernel_name = os.environ.get('IR_KERNEL_NAME', 'testir')

    language_name = 'R'

    def _execute_code(self, code, tests=True, silent=False, store_history=True):
        self.flush_channels()

        reply, output_msgs = self.execute_helper(code, silent=silent, store_history=store_history)

        self.assertEqual(reply['content']['status'], 'ok', '{0}: {0}'.format(reply['content'].get('ename'), reply['content'].get('evalue')))
        if tests:
            self.assertGreaterEqual(len(output_msgs), 1)
            # the irkernel only sends display_data, not execute_results
            self.assertEqual(output_msgs[0]['msg_type'], 'display_data')
        return reply, output_msgs

    code_hello_world = 'print("hello, world")'

    completion_samples = [
        {'text': 'zi',                     'matches': {'zip'}},
        {'text': 'seq_len(',               'matches': {'length.out = '}},
        {'text': 'base::transform(',       'matches': {'`_data` = ', '...'}},
        {'text': 'foo(R_system',           'matches': {'R_system_version'}},
        {'text': 'version$plat',           'matches': {'version$platform'}},
        {'text': 'stats4::AIC@def',        'matches': {'stats4::AIC@default'}},
        {'text': 'stats4::AIC@default@ta', 'matches': {'stats4::AIC@default@target'}},
        {'text': 'grDevice',               'matches': {'grDevices::'}},
        {'text': 'base::abbrev',           'matches': {'base::abbreviate'}},
        {'text': 'base::.rowNamesD',       'matches': {'base::`.rowNamesDF<-`'}},
        {'text': 'repr:::repr_png.def',    'matches': {'repr:::repr_png.default'}},
        {'text': 'repr::format2repr$mark', 'matches': {'repr::format2repr$markdown'}},
    ]

    complete_code_samples = ['1', 'print("hello, world")', 'f <- function(x) {\n  x*2\n}']

    incomplete_code_samples = ['print("hello', 'f <- function(x) {\n  x*2']

    code_display_data = [
        {'code': '"a"', 'mime': 'text/plain'},
        {'code': '"a"', 'mime': 'text/html'},
        {'code': '"a"', 'mime': 'text/latex'},
        {'code': '1:3', 'mime': 'text/plain'},
        {'code': '1:3', 'mime': 'text/html'},
        {'code': '1:3', 'mime': 'text/latex'},
        {'code': without_rich_display.format('"a"'), 'mime': 'text/plain'},
        {'code': without_rich_display.format('1:3'), 'mime': 'text/plain'},
    ]

    def test_display_vector(self):
        """display of vectors"""
        code = '1:3'
        reply, output_msgs = self._execute_code(code)
        
        # we currently send those formats: text/plain, text/html, text/latex, and text/markdown
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 4, data.keys())
        self.assertEqual(data['text/plain'], '[1] 1 2 3')
        self.assertIn('text/html', data)
        
        # this should not be a test of the repr functionality...
        self.assertIn('</li>', output_msgs[0]['content']['data']['text/html'])
        self.assertIn('text/latex', output_msgs[0]['content']['data'])
        self.assertIn('text/markdown', output_msgs[0]['content']['data'])

    def test_display_vector_only_plaintext(self):
        """display of plain text vectors"""
        code = without_rich_display.format('1:3')
        reply, output_msgs = self._execute_code(code)
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 1, data.keys())
        self.assertEqual(data['text/plain'], '[1] 1 2 3')

    def test_irkernel_plots(self):
        """plotting"""
        code = 'plot(1:3)'
        reply, output_msgs = self._execute_code(code)
        
        # we currently send two formats: png, and text/plain
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 2, data.keys())
        self.assertEqual(data['text/plain'], 'plot without title')
        self.assertIn('image/png', data)
        
        # we isolate only svg plots, which are not included in the default types
        metadata = output_msgs[0]['content']['metadata']
        self.assertEqual(len(metadata), 0, metadata.keys())

    def test_irkernel_plots_only_PNG(self):
        """plotting PNG"""
        # the reset needs to happen in another execute because plots are sent after either
        # the next plot is opened or everything is executed, not at the time when plot
        # command is actually happening.
        # (we have a similar problem with width/... but we worked around it by setting the
        # appropriate options to the recorderdplot object)
        code = '''\
            old_options <- options(jupyter.plot_mimetypes = c('image/png'))
            plot(1:3)
        '''
        reply, output_msgs = self._execute_code(code)
        
        # Only png, no svg or plain/text
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 1, data.keys())
        self.assertIn('image/png', data)

        # nothing in metadata
        metadata = output_msgs[0]['content']['metadata']
        self.assertEqual(len(metadata), 0, metadata.keys())

        # And reset
        code = 'options(old_options)'
        reply, output_msgs = self._execute_code(code, tests=False)

    def test_irkernel_plots_only_SVG(self):
        # again the reset dance (see PNG)
        code = '''\
            old_options <- options(jupyter.plot_mimetypes = c('image/svg+xml'))
            plot(1:3)
        '''
        reply, output_msgs = self._execute_code(code)

        # Only svg, no png or plain/text
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 1, data.keys())
        self.assertIn('image/svg+xml', data)
        
        # svg output is currently isolated
        metadata = output_msgs[0]['content']['metadata']
        self.assertEqual(len(metadata), 1, metadata.keys())
        self.assertEqual(len(metadata['image/svg+xml']), 1)
        self.assertEqual(metadata['image/svg+xml']['isolated'], True)

        # And reset
        code = 'options(old_options)'
        reply, output_msgs = self._execute_code(code, tests=False)

    def test_irkernel_plots_without_rich_display(self):
        code = '''\
            options(jupyter.rich_display = FALSE)
            plot(1:3)
        '''
        reply, output_msgs = self._execute_code(code)

        # Even with rich output as false, we send plots
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 2, data.keys())
        self.assertEqual(data['text/plain'], 'plot without title')
        self.assertIn('image/png', data)

        # And reset
        code = 'options(jupyter.rich_display = TRUE)'
        reply, output_msgs = self._execute_code(code, tests=False)

    def test_irkernel_df_default_rich_output(self):
        """data.frame rich representation"""
        code = 'data.frame(x = 1:3)'
        reply, output_msgs = self._execute_code(code)
        
        # we currently send three formats: text/plain, html, and latex
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 4, data.keys())

    def test_irkernel_df_no_rich_output(self):
        """data.frame plain representation"""
        code = '''
            options(jupyter.rich_display = FALSE)
            data.frame(x = 1:3)
            options(jupyter.rich_display = TRUE)
        '''
        reply, output_msgs = self._execute_code(code)
        
        # only text/plain
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 1, data.keys())

    def test_html_isolated(self):
        """HTML isolation"""
        code = '''
            repr_html.full_page <- function(obj) sprintf('<html><body>%s</body></html>', obj)
            structure(0, class = 'full_page')
        '''
        reply, output_msgs = self._execute_code(code)
        
        data = output_msgs[0]['content']['data']
        self.assertEqual(len(data), 2, data.keys())
        self.assertEqual(data['text/html'], '<html><body>0</body></html>')
        
        metadata = output_msgs[0]['content']['metadata']
        self.assertEqual(len(metadata), 1, metadata.keys())
        self.assertEqual(len(metadata['text/html']), 1, metadata['text/html'].keys())
        self.assertEqual(metadata['text/html']['isolated'], True)

    def test_in_kernel_set(self):
        """jupyter.in_kernel option"""
        reply, output_msgs = self._execute_code('getOption("jupyter.in_kernel")')
        data = output_msgs[0]['content']['data']
        self.assertGreaterEqual(len(data), 1, data.keys())
        self.assertEqual(data['text/plain'], '[1] TRUE', data.keys())
    
    def test_warning_message(self):
        self.flush_channels()
        reply, output_msgs = self.execute_helper('options(warn=1); warning(simpleWarning("wmsg"))')
        self.assertEqual(output_msgs[0]['msg_type'], 'stream')
        self.assertEqual(output_msgs[0]['content']['name'], 'stderr')
        self.assertEqual(output_msgs[0]['content']['text'], 'Warning message:\n“wmsg”')
        
        self.flush_channels()
        reply, output_msgs = self.execute_helper('options(warn=1); f <- function() warning("wmsg"); f()')
        self.assertEqual(output_msgs[0]['msg_type'], 'stream')
        self.assertEqual(output_msgs[0]['content']['name'], 'stderr')
        self.assertEqual(output_msgs[0]['content']['text'], 'Warning message in f():\n“wmsg”')

    def test_should_increment_history(self):
        """properly increments execution history"""
        code = 'data.frame(x = 1:3)'
        reply, output_msgs = self._execute_code(code)
        reply2, output_msgs2 = self._execute_code(code)
        execution_count_1 = reply['content']['execution_count']
        execution_count_2 = reply2['content']['execution_count']
        self.assertEqual(execution_count_1 + 1, execution_count_2)

    def test_should_not_increment_history(self):
        """Does not increment history if silent is true or store_history is false"""
        code = 'data.frame(x = 1:3)'
        reply, output_msgs = self._execute_code(code, store_history=False)
        reply2, output_msgs2 = self._execute_code(code, store_history=False)
        reply3, output_msgs3 = self._execute_code(code, tests=False, silent=True)
        execution_count_1 = reply['content']['execution_count']
        execution_count_2 = reply2['content']['execution_count']
        execution_count_3 = reply3['content']['execution_count']
        self.assertEqual(execution_count_1, execution_count_2)
        self.assertEqual(execution_count_1, execution_count_3)

    def test_irkernel_inspects(self):
        """Test if object inspection works."""
        self.flush_channels()

        def test_token_is_ok(token, preprocess=None, postprocess=None):
            """Check if inspect_request for the `token` returns a reply.

            Run code in `preprocess` before requesting if it's given,
            and `proprocess` after requesting.

            Currently just test if the kernel replys without an error
            and not care about its content.
            Because the contents of inspections are still so arguable.
            When the requirements for the contents are decided,
            fix the tests beow and check the contents.
            """
            if preprocess:
                self._execute_code(preprocess, tests=False)

            msg_id = self.kc.inspect(token)
            reply = self.kc.get_shell_msg(timeout=TIMEOUT)
            validate_message(reply, 'inspect_reply', msg_id)

            self.assertEqual(reply['content']['status'], 'ok')
            self.assertTrue(reply['content']['found'])
            self.assertGreaterEqual(len(reply['content']['data']), 1)

            if postprocess:
                self._execute_code(postprocess, tests=False)

        # Numeric constant
        test_token_is_ok('1')
        # Reserved word
        test_token_is_ok('NULL')
        # Dataset with a help document
        test_token_is_ok('iris')
        # Function with a help document
        test_token_is_ok('c')
        # Function name with namespace
        test_token_is_ok('base::c')
        # Function not exported from namespace
        test_token_is_ok('tools:::.Rd2pdf')
        # User-defined variable
        test_token_is_ok(
            'x',
            preprocess='x <- 1',
            postprocess='rm("x")'
        )
        # User-defined function
        test_token_is_ok(
            'f',
            preprocess='f <- function (x) x + x',
            postprocess='rm("f")'
        )
        # Object which masks other object in workspace
        test_token_is_ok(
            'c',
            preprocess='c <- function (x) x + x',
            postprocess='rm("c")'
        )


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Machinery for testing Jupyter kernels via the messaging protocol.
"""

from unittest import TestCase, SkipTest
try:                  # Python 3
    from queue import Empty
except ImportError:   # Python 2
    from Queue import Empty

from jupyter_client.manager import start_new_kernel
from .messagespec import validate_message, MimeBundle

TIMEOUT = 15

__version__ = '0.2.2'

class KernelTests(TestCase):
    kernel_name = ""

    @classmethod
    def setUpClass(cls):
        cls.km, cls.kc = start_new_kernel(kernel_name=cls.kernel_name)

    @classmethod
    def tearDownClass(cls):
        cls.kc.stop_channels()
        cls.km.shutdown_kernel()

    def flush_channels(self):
        for channel in (self.kc.shell_channel, self.kc.iopub_channel):
            while True:
                try:
                    msg = channel.get_msg(block=True, timeout=0.1)
                except Empty:
                    break
                else:
                    validate_message(msg)

    language_name = ""
    file_extension = ""

    def test_kernel_info(self):
        self.flush_channels()

        msg_id = self.kc.kernel_info()
        reply = self.kc.get_shell_msg(timeout=TIMEOUT)
        validate_message(reply, 'kernel_info_reply', msg_id)

        if self.language_name:
            self.assertEqual(reply['content']['language_info']['name'],
                             self.language_name)
        if self.file_extension:
            self.assertEqual(reply['content']['language_info']['file_extension'],
                             self.file_extension)
            self.assertTrue(reply['content']['language_info']['file_extension'].startswith("."))



    def execute_helper(self, code, timeout=TIMEOUT,
                       silent=False, store_history=True):
        msg_id = self.kc.execute(code=code, silent=silent,
                                 store_history=store_history)

        reply = self.kc.get_shell_msg(timeout=timeout)
        validate_message(reply, 'execute_reply', msg_id)

        busy_msg = self.kc.iopub_channel.get_msg(timeout=1)
        validate_message(busy_msg, 'status', msg_id)
        self.assertEqual(busy_msg['content']['execution_state'], 'busy')

        output_msgs = []
        while True:
            msg = self.kc.iopub_channel.get_msg(timeout=0.1)
            validate_message(msg, msg['msg_type'], msg_id)
            if msg['msg_type'] == 'status':
                self.assertEqual(msg['content']['execution_state'], 'idle')
                break
            elif msg['msg_type'] == 'execute_input':
                self.assertEqual(msg['content']['code'], code)
                continue
            output_msgs.append(msg)

        return reply, output_msgs

    code_hello_world = ""

    def test_execute_stdout(self):
        if not self.code_hello_world:
            raise SkipTest

        self.flush_channels()
        reply, output_msgs = self.execute_helper(code=self.code_hello_world)

        self.assertEqual(reply['content']['status'], 'ok')

        self.assertGreaterEqual(len(output_msgs), 1)
        self.assertEqual(output_msgs[0]['msg_type'], 'stream')
        self.assertEqual(output_msgs[0]['content']['name'], 'stdout')
        self.assertIn('hello, world', output_msgs[0]['content']['text'])

    code_stderr = ""

    def test_execute_stderr(self):
        if not self.code_stderr:
            raise SkipTest

        self.flush_channels()
        reply, output_msgs = self.execute_helper(code=self.code_stderr)

        self.assertEqual(reply['content']['status'], 'ok')

        self.assertGreaterEqual(len(output_msgs), 1)
        self.assertEqual(output_msgs[0]['msg_type'], 'stream')
        self.assertEqual(output_msgs[0]['content']['name'], 'stderr')

    completion_samples = []

    def test_completion(self):
        if not self.completion_samples:
            raise SkipTest

        for sample in self.completion_samples:
            with self.subTest(text=sample['text']):
                msg_id = self.kc.complete(sample['text'])
                reply = self.kc.get_shell_msg()
                validate_message(reply, 'complete_reply', msg_id)
                if 'matches' in sample:
                    self.assertEqual(set(reply['content']['matches']),
                                     set(sample['matches']))

    complete_code_samples = []
    incomplete_code_samples = []
    invalid_code_samples = []

    def check_is_complete(self, sample, status):
        msg_id = self.kc.is_complete(sample)
        reply = self.kc.get_shell_msg()
        validate_message(reply, 'is_complete_reply', msg_id)
        if reply['content']['status'] != status:
            msg = "For code sample\n  {!r}\nExpected {!r}, got {!r}."
            raise AssertionError(msg.format(sample, status,
                                            reply['content']['status']))

    def test_is_complete(self):
        if not (self.complete_code_samples
                or self.incomplete_code_samples
                or self.invalid_code_samples):
            raise SkipTest

        self.flush_channels()

        with self.subTest(status="complete"):
            for sample in self.complete_code_samples:
                self.check_is_complete(sample, 'complete')

        with self.subTest(status="incomplete"):
            for sample in self.incomplete_code_samples:
                self.check_is_complete(sample, 'incomplete')

        with self.subTest(status="invalid"):
            for sample in self.invalid_code_samples:
                self.check_is_complete(sample, 'invalid')

    code_page_something = ""

    def test_pager(self):
        if not self.code_page_something:
            raise SkipTest

        self.flush_channels()

        reply, output_msgs = self.execute_helper(self.code_page_something)
        self.assertEqual(reply['content']['status'],  'ok')
        payloads = reply['content']['payload']
        self.assertEqual(len(payloads), 1)
        self.assertEqual(payloads[0]['source'], 'page')
        mimebundle = payloads[0]['data']
        # Validate the mimebundle
        MimeBundle().data = mimebundle
        self.assertIn('text/plain', mimebundle)

    code_generate_error = ""

    def test_error(self):
        if not self.code_generate_error:
            raise SkipTest

        self.flush_channels()

        reply, output_msgs = self.execute_helper(self.code_generate_error)
        self.assertEqual(reply['content']['status'], 'error')
        self.assertEqual(len(output_msgs), 1)
        self.assertEqual(output_msgs[0]['msg_type'], 'error')

    code_execute_result = []

    def test_execute_result(self):
        if not self.code_execute_result:
            raise SkipTest

        for sample in self.code_execute_result:
            with self.subTest(code=sample['code']):
                self.flush_channels()

                reply, output_msgs = self.execute_helper(sample['code'])

                self.assertEqual(reply['content']['status'], 'ok')

                self.assertGreaterEqual(len(output_msgs), 1)
                self.assertEqual(output_msgs[0]['msg_type'], 'execute_result')
                self.assertIn('text/plain', output_msgs[0]['content']['data'])
                self.assertEqual(output_msgs[0]['content']['data']['text/plain'],
                                 sample['result'])

    code_display_data = []

    def test_display_data(self):
        if not self.code_display_data:
            raise SkipTest

        for sample in self.code_display_data:
            with self.subTest(code=sample['code']):
                self.flush_channels()
                reply, output_msgs = self.execute_helper(sample['code'])

                self.assertEqual(reply['content']['status'], 'ok')

                self.assertGreaterEqual(len(output_msgs), 1)
                self.assertEqual(output_msgs[0]['msg_type'], 'display_data')
                self.assertIn(sample['mime'], output_msgs[0]['content']['data'])

    # this should match one of the values in code_execute_result
    code_history_pattern = ""
    supported_history_operations = ()

    def history_helper(self, execute_first, timeout=TIMEOUT, **histargs):
        self.flush_channels()

        for code in execute_first:
            reply, output_msgs = self.execute_helper(code)

        self.flush_channels()
        msg_id = self.kc.history(**histargs)

        reply = self.kc.get_shell_msg(timeout=timeout)
        validate_message(reply, 'history_reply', msg_id)

        return reply

    def test_history(self):
        if not self.code_execute_result:
            raise SkipTest

        codes = [s['code'] for s in self.code_execute_result]
        results = [s['result'] for s in self.code_execute_result]
        n = len(codes)

        session = start = None

        with self.subTest(hist_access_type="tail"):
            if 'tail' not in self.supported_history_operations:
                raise SkipTest
            reply = self.history_helper(codes, output=False, raw=True,
                                        hist_access_type="tail", n=n)
            self.assertEqual(len(reply['content']['history']), n)
            self.assertEqual(len(reply['content']['history'][0]), 3)
            self.assertEqual(codes, [h[2] for h in reply['content']['history']])

            session, start = reply['content']['history'][0][0:2]
            with self.subTest(output=True):
                reply = self.history_helper(codes, output=True, raw=True,
                                            hist_access_type="tail", n=n)
                self.assertEqual(len(reply['content']['history'][0][2]), 2)

        with self.subTest(hist_access_type="range"):
            if 'range' not in self.supported_history_operations:
                raise SkipTest
            if session is None:
                raise SkipTest
            reply = self.history_helper(codes, output=False, raw=True,
                                        hist_access_type="range",
                                        session=session, start=start,
                                        stop=start+1)
            self.assertEqual(len(reply['content']['history']), 1)
            self.assertEqual(reply['content']['history'][0][0], session)
            self.assertEqual(reply['content']['history'][0][1], start)

        with self.subTest(hist_access_type="search"):
            if not self.code_history_pattern:
                raise SkipTest
            if 'search' not in self.supported_history_operations:
                raise SkipTest

            with self.subTest(subsearch="normal"):
                reply = self.history_helper(codes, output=False, raw=True,
                                            hist_access_type="search",
                                            pattern=self.code_history_pattern)
                self.assertGreaterEqual(len(reply['content']['history']), 1)
            with self.subTest(subsearch="unique"):
                reply = self.history_helper(codes, output=False, raw=True,
                                            hist_access_type="search",
                                            pattern=self.code_history_pattern,
                                            unique=True)
                self.assertEqual(len(reply['content']['history']), 1)
            with self.subTest(subsearch="n"):
                reply = self.history_helper(codes, output=False, raw=True,
                                            hist_access_type="search",
                                            pattern=self.code_history_pattern,
                                            n=3)
                self.assertEqual(len(reply['content']['history']), 3)

    code_inspect_sample = ""

    def test_inspect(self):
        if not self.code_inspect_sample:
            raise SkipTest

        self.flush_channels()
        msg_id = self.kc.inspect(self.code_inspect_sample)
        reply = self.kc.get_shell_msg(timeout=TIMEOUT)
        validate_message(reply, 'inspect_reply', msg_id)

        self.assertEqual(reply['content']['status'], 'ok')
        self.assertTrue(reply['content']['found'])
        self.assertGreaterEqual(len(reply['content']['data']), 1)

    code_clear_output = ""

    def test_clear_output(self):
        if not self.code_clear_output:
            raise SkipTest

        self.flush_channels()
        reply, output_msgs = self.execute_helper(code=self.code_clear_output)

        self.assertEqual(reply['content']['status'], 'ok')

        self.assertGreaterEqual(len(output_msgs), 1)
        self.assertEqual(output_msgs[0]['msg_type'], 'clear_output')



"""The v5 message specification. This can be used in the tests to verify messages."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import sys

import nose.tools as nt
from traitlets import (
    Bool, Unicode, Dict, Integer, List, Enum,
)

from .messagespec_common import Reference, MimeBundle, Version

# message and header formats specific to verion 5

class RMessage(Reference):
    msg_id = Unicode()
    msg_type = Unicode()
    header = Dict()
    parent_header = Dict()
    content = Dict()
    
    def check(self, d):
        super(RMessage, self).check(d)
        RHeader().check(self.header)
        if self.parent_header:
            RHeader().check(self.parent_header)

class RHeader(Reference):
    msg_id = Unicode()
    msg_type = Unicode()
    session = Unicode()
    username = Unicode()
    version = Version(min='5.0')

# version shell replies

class ExecuteReply(Reference):
    execution_count = Integer()
    status = Enum((u'ok', u'error'))
    
    def check(self, d):
        Reference.check(self, d)
        if d['status'] == 'ok':
            ExecuteReplyOkay().check(d)
        elif d['status'] == 'error':
            ExecuteReplyError().check(d)


class ExecuteReplyOkay(Reference):
    payload = List(Dict)
    user_expressions = Dict()


class ExecuteReplyError(Reference):
    ename = Unicode()
    evalue = Unicode()
    traceback = List(Unicode)


class InspectReply(MimeBundle):
    found = Bool()


class ArgSpec(Reference):
    args = List(Unicode)
    varargs = Unicode()
    varkw = Unicode()
    defaults = List()


class Status(Reference):
    execution_state = Enum((u'busy', u'idle', u'starting'))


class CompleteReply(Reference):
    matches = List(Unicode)
    cursor_start = Integer()
    cursor_end = Integer()
    status = Unicode()

class LanguageInfo(Reference):
    name = Unicode()
    version = Unicode(sys.version.split()[0])

class KernelInfoReply(Reference):
    protocol_version = Version(min='5.0')
    implementation = Unicode()
    implementation_version = Version()
    language_info = Dict()
    banner = Unicode()
    
    def check(self, d):
        Reference.check(self, d)
        LanguageInfo().check(d['language_info'])


class IsCompleteReply(Reference):
    status = Enum((u'complete', u'incomplete', u'invalid', u'unknown'))
    
    def check(self, d):
        Reference.check(self, d)
        if d['status'] == 'incomplete':
            IsCompleteReplyIncomplete().check(d)

class IsCompleteReplyIncomplete(Reference):
    indent = Unicode()


# IOPub messages

class ExecuteInput(Reference):
    code = Unicode()
    execution_count = Integer()


Error = ExecuteReplyError


class Stream(Reference):
    name = Enum((u'stdout', u'stderr'))
    text = Unicode()


class DisplayData(MimeBundle):
    pass


class ExecuteResult(MimeBundle):
    execution_count = Integer()

class HistoryReply(Reference):
    history = List(List())

class ClearOutput(Reference):
    wait = Bool()


"""
Specifications of `content` part of the reply messages.
"""
references = {
    'execute_reply' : ExecuteReply(),
    'inspect_reply' : InspectReply(),
    'status' : Status(),
    'complete_reply' : CompleteReply(),
    'kernel_info_reply': KernelInfoReply(),
    'is_complete_reply': IsCompleteReply(),
    'execute_input' : ExecuteInput(),
    'execute_result' : ExecuteResult(),
    'history_reply' : HistoryReply(),
    'error' : Error(),
    'stream' : Stream(),
    'display_data' : DisplayData(),
    'header' : RHeader(),
    'clear_output' : ClearOutput(),
}

#validation specific to this version of the message specification

def validate_message(msg, msg_type=None, parent=None):
    """validate a message
    
    If msg_type and/or parent are given, the msg_type and/or parent msg_id
    are compared with the given values.
    """
    RMessage().check(msg)
    if msg_type:
        nt.assert_equal(msg['msg_type'], msg_type)
    if parent:
        nt.assert_equal(msg['parent_header']['msg_id'], parent)
    content = msg['content']
    ref = references[msg['msg_type']]
    ref.check(content)
